import time
from typing import Callable, Generic, TypeVar
from .delegate import action
from .readonly import readonly


T = TypeVar("T")


class RxEvent(Generic[T]):
    def __init__(self, pre: T, post: T) -> None:
        self.__pre = pre
        self.__post = post

    @property
    def pre(self) -> T:
        return self.__pre

    @property
    def new(self) -> T:
        return self.__post


class ReactiveProperty(Generic[T]):
    def __init__(self, value: T) -> None:
        self.__value = value
        self.on_changed = action[RxEvent[T]]()

    @property
    def value(self) -> T:
        return self.__value

    @value.setter
    def value(self, new_value: T) -> None:
        if self.__value != new_value:
            self.__value = new_value
            self.on_changed.invoke(RxEvent(self.__value, new_value))

    def subscribe(
        self,
        cb: Callable[[RxEvent[T]], None],
        immediate: bool = False,
    ) -> None:
        self.on_changed.register(cb)
        if immediate:
            self.notify()

    def notify(self) -> None:
        self.on_changed.invoke(RxEvent(self.__value, self.__value))

    def as_readonly(self) -> "readonly[T]":
        return readonly(lambda: self.__value)

    def dispose(self) -> None:
        self.on_changed.clear()


if __name__ == "__main__":
    name = ReactiveProperty[str]("name0")
    name.subscribe(lambda e: print("changed from", e.pre, "to", e.new))

    for i in range(1, 10 + 1):
        name.value = f"name{i}"
        time.sleep(0.2)



from typing import Generic, Optional, TypeVar


T = TypeVar("T")


class nullable(Generic[T]):
    def __init__(self, value: Optional[T]) -> None:
        self.__value = value

    @property
    def value(self) -> T:
        if self.__value:
            return self.__value
        raise ValueError("No value present")

    @value.setter
    def value(self, value: Optional[T]) -> None:
        self.__value = value

    @property
    def has_value(self) -> bool:
        return self.__value is not None


def as_nullable(value: T) -> nullable[T]:
    return nullable(value)


if __name__ == "__main__":
    name = nullable[str](None)
    print(name.has_value, "\t:", name.value if name.has_value else "-")
    name.value = "a"
    print(name.has_value, "\t:", name.value if name.has_value else "-")



from typing import Callable, Generic, Iterator, TypeVar, TypeVarTuple


T_RESULT = TypeVar("T_RESULT")
T_SOURCE = TypeVarTuple("T_SOURCE")


class action(Generic[*T_SOURCE]):
    def __init__(self, *callbacks: Callable[[*T_SOURCE], None]) -> None:
        self.__callbacks = list(callbacks)

    def register(self, *callbacks: Callable[[*T_SOURCE], None]) -> None:
        for cb in callbacks:
            self.__callbacks.append(cb)

    def unregister(self, *callbacks: Callable[[*T_SOURCE], None]) -> None:
        for cb in callbacks:
            self.__callbacks.remove(cb)

    def invoke(self, *args: *T_SOURCE) -> None:
        if self.__callbacks:
            for cb in self.__callbacks:
                cb(*args)

    def as_callable(self) -> Iterator[Callable[[*T_SOURCE], None]]:
        for cb in self.__callbacks:
            yield cb

    def clear(self) -> None:
        self.__callbacks.clear()

    def __iadd__(self, callback: Callable[[*T_SOURCE], None]) -> "action[*T_SOURCE]":
        self.__callbacks.append(callback)
        return self

    def __isub__(self, callback: Callable[[*T_SOURCE], None]) -> "action[*T_SOURCE]":
        self.__callbacks.remove(callback)
        return self

    def __len__(self) -> int:
        return len(self.__callbacks)

    def __call__(self, *args: *T_SOURCE) -> None:
        for cb in self.__callbacks:
            cb(*args)


class func(Generic[*T_SOURCE, T_RESULT]):
    def __init__(self, *callbacks: Callable[[*T_SOURCE], T_RESULT]) -> None:
        self.__callbacks = list(callbacks)

    def register(self, *callbacks: Callable[[*T_SOURCE], T_RESULT]) -> None:
        for cb in callbacks:
            self.__callbacks.append(cb)

    def unregister(self, *callbacks: Callable[[*T_SOURCE], T_RESULT]) -> None:
        for cb in callbacks:
            self.__callbacks.remove(cb)

    def invoke(self, *args: *T_SOURCE) -> T_RESULT:
        if self.__callbacks.__len__() <= 0:
            raise ValueError("No callbacks registered")
        return [cb(*args) for cb in self.__callbacks][0]

    def as_callable(self) -> Callable[[*T_SOURCE], T_RESULT]:
        return self.__call__

    def clear(self) -> None:
        self.__callbacks.clear()

    def __iadd__(
        self, callback: Callable[[*T_SOURCE], T_RESULT]
    ) -> "func[*T_SOURCE, T_RESULT]":
        self.__callbacks.append(callback)
        return self

    def __isub__(
        self, callback: Callable[[*T_SOURCE], T_RESULT]
    ) -> "func[*T_SOURCE, T_RESULT]":
        self.__callbacks.remove(callback)
        return self

    def __len__(self) -> int:
        return len(self.__callbacks)

    def __call__(self, *args: *T_SOURCE) -> T_RESULT:
        return [cb(*args) for cb in self.__callbacks][0]


def on_int_value_changed(i: int) -> None:
    print("on_int_value_changed:", i)


def on_authorised_callback() -> bool:
    print("on_authorised")
    return True


if __name__ == "__main__":
    on_value_changed = action[int]()
    on_value_changed += on_int_value_changed
    on_value_changed += on_int_value_changed
    on_value_changed -= on_int_value_changed
    on_value_changed(12)

    on_authorised = func[bool]()
    on_authorised += on_authorised_callback
    on_authorised += on_authorised_callback
    on_authorised -= on_authorised_callback
    print("authorise result:", on_authorised())

    on_user_updated = action[int, float, str]()
    on_user_updated += lambda user_id, score, name: print(user_id, score, name)
    on_user_updated.invoke(123, 99.8, "trrne")



from typing import Callable, Generic, TypeVar
from .delegate import func

T = TypeVar("T")


class readonly(Generic[T]):
    def __init__(self, getter: Callable[[], T] | func[T]):
        self.__getter = func[T](getter)

    @property
    def value(self) -> T:
        return self.__getter.invoke()


def as_readonly(obj: T) -> readonly[T]:
    return readonly(lambda: obj)



from setuptools import setup

setup(
    name="omnim",
    url="https://github.com/cet-t/omnipy",
    version="0.1.1",
    description="useful toolkit",
    author="cet",
    packages=["omnim"],
    include_package_data=True,
    zip_safe=True,
)
